# Pyflow an open-source tool for modular visual programing in python
# Copyright (C) 2021-2022 Bycelium <https://www.gnu.org/licenses/>

""" Module for the  Blocks Widgets. """

from pyflow.blocks.widgets.blocksplitter import Splitter
from pyflow.blocks.widgets.blocktitle import Title
from pyflow.blocks.widgets.blocksizegrip import SizeGrip
