# Pyflow an open-source tool for modular visual programing in python
# Copyright (C) 2021-2022 Bycelium <https://www.gnu.org/licenses/>

""" Module for the Scene creation and manipulations."""

from pyflow.scene.scene import Scene
