# Pyflow an open-source tool for modular visual programing in python
# Copyright (C) 2021-2022 Bycelium <https://www.gnu.org/licenses/>

""" Module for the  Blocks of different types. """

from pyflow.blocks.sliderblock import SliderBlock
from pyflow.blocks.codeblock import CodeBlock
from pyflow.blocks.markdownblock import MarkdownBlock
from pyflow.blocks.drawingblock import DrawingBlock
from pyflow.blocks.containerblock import ContainerBlock
